#include <algorithm>
#include <cassert>
#include <cstddef>
#include <cstdlib>
#include <fstream>
#include <iostream>
#include <ostream>
#include <numeric>
#include <sstream>
#include <string>
#include <utility>
#include <vector>

std::string bwt(std::string text, char special = '$')
{
}

std::string inverse_bwt(std::string text, char special = '$')
{
    typedef std::pair<std::size_t, char> indexed_char;
    std::vector<indexed_char> tokens(text.length());

    for (std::size_t i = 0; i < text.length(); ++i)
    {
        tokens[i] = std::make_pair(i, text[i]);
    }

    std::stable_sort(tokens.begin(), tokens.end(),
        [special](indexed_char a, indexed_char b)
        {
            if (a.second == special)
            {
                return false;
            }

            if (b.second == special)
            {
                return true;
            }

            return a.second < b.second;
        }
        );

    auto it = std::find_if(tokens.begin(), tokens.end(),
        [](indexed_char a)
        {
            return a.second == '$';
        }
        );

    assert(it != tokens.end());

    std::size_t first = (*it).first;

    std::string result;
    for (std::size_t i = 0; i < text.length(); ++i)
    {
        char c = tokens[first].second;
        first = tokens[first].first;
        result.append(1, c);
    }


    return result;
}

int main(int argc, char** argv)
{
#if 1
    if (argc != 2)
    {
        std::cerr << "Usage: " << argv[0] << ' ' << "<textfile>\n";
        return EXIT_FAILURE;
    }

    // Read textfile
    std::ifstream textfile(argv[1]);
    if (!textfile.good())
    {
        std::cerr << "Error reading from file " << argv[1] << '\n';
        return EXIT_FAILURE;
    }

    std::stringstream stream;
    stream << textfile.rdbuf();

    // Ascertain there's no '$'
    std::istream_iterator<char> begin(stream);
    std::istream_iterator<char> end;
    if (std::find(begin, end, '$') != end)
    {
        std::cerr << "File may not contain special character '$'\n";
        return EXIT_FAILURE;
    }

    std::string text = stream.str();
    text.append(1, '$');
#else
    // Some simple tests
    std::string text = "ANANAS$";
//  std::string text = "BANANA$";
//  std::string text = "EINSZWEI$";
//  std::string text = "MISSISSIPPI$";
//  std::string text = "PINEAPPLE$";
#endif

    // Print original text
    std::cout << text << '\n';

    // Transform and print
    text = bwt(text);
    std::cout << text << '\n';

    // Inverse transform
    // should then print the original text again!
    text = inverse_bwt(text);
    std::cout << text << '\n';

    return EXIT_SUCCESS;
}
